/**
 * Filewatch watches the directories specified at comandline (recursively) for 
 * changes and mails them when either 1000 lines have been gathered or 10 
 * seconds passed without out any change.
 *
 * Copyright (C) 2008  Niko Efthymiou <nefthy-filewatch@nefthy.de>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; version 2
 * of the License and no later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, 
 * USA.
 *
 *
 * BUGS: Killing the program at comandline through ^C also kills the spawned
 * mta, which may cause loss of data. To fix this messages could be buffered
 * and send to the mta when flush() is called (which also happend before program
 * termination, unless it is killed with SIGKILL)
 */
#define _GNU_SOURCE
#include <signal.h>
#include <stdlib.h>
#include <errno.h>
#include <time.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/inotify.h>

#define MAIL_MAX_LINES 1000
#define MAIL_INACTIVITY_TIMEOUT 10
#define MAIL_TO "nefthy"
#define MAIL_PROG "qmail-inject"

char** dirs = NULL;
size_t dirs_len = 0;
volatile sig_atomic_t flush = 0;
volatile sig_atomic_t done = 0;
FILE* mail = NULL;
int mail_lines = 0;

void out_flush() {
	puts("flushing");
	if (NULL != mail) {
		int exit_code = pclose(mail);
		if (exit_code < 0) {
			perror("pclose failed");
		} else if (exit_code != 0) {
			printf("qmail exited with %d\n", exit_code);
		}
	}
	mail_lines = 0;
	mail = popen(MAIL_PROG, "w");
	if (NULL == mail) {
		perror("popen failed");
	}
	fputs("To: " MAIL_TO "\nSubject: filesystem changes\n\n", mail);
}

void out_append(char* s) {
#if DEBUG
	fputs(s, stdout);
#endif
	fputs(s, mail);
}

void out_end_line() {
#if DEBUG
	fputs("\n", stdout);
#endif
	fputs("\n", mail);
	mail_lines++;
	if (mail_lines > MAIL_MAX_LINES) {
		out_flush();
	}
}
void add(int notify, char* dirname) {
	const uint32_t mask = IN_ATTRIB | IN_CREATE | IN_DELETE | IN_DELETE_SELF 
		| IN_MODIFY | IN_MOVE_SELF | IN_MOVED_FROM	| IN_MOVED_TO;
	
	int watch = inotify_add_watch(notify, dirname, mask);
	printf("%s has wd %d\n", dirname, watch);
	if (-1 == watch) {
		perror("failed to add watch");
		return;
	}
	if (dirs_len < (size_t) watch) {
		char** tmp = realloc(dirs, sizeof(char *) * (watch + 10));
		if (NULL == tmp) {
			perror("realloc failed");
			inotify_rm_watch(notify, watch);
			return;
		}
		dirs = tmp;
	}

	dirs[watch-1] = strdup(dirname);
}

void rm(int notify, uint32_t watch) {
	if (0 != inotify_rm_watch(notify, watch)) {
		perror("failed to remove watch");
	}
	free(dirs[watch-1]);
}

void add_r(int notify, char* dirname) {
	struct stat st;
	struct dirent *dt;
	char filename[strlen(dirname)+1024+1]; //XXX: BUG! 1024 falls out of the blue
	stat(dirname, &st);
	if (S_ISDIR(st.st_mode)) {
		DIR* d = opendir(dirname);
		if (NULL == d) {
			perror("diropen failed");
			errno = 0;
			return;
		}
		while (NULL != (dt = readdir(d))) {
			if (0 == strcmp(dt->d_name, ".") || 0 == strcmp(dt->d_name, "..")) {
				continue;
			}

			if (DT_DIR == dt->d_type) {
				snprintf(filename, sizeof(filename), "%s/%s", dirname, dt->d_name); 
				//add(notify, filename);
				add_r(notify, filename);
			}
			//free(dt);
		}
		closedir(d);
		add(notify, dirname);
	} else {
		fprintf(stderr, "Ignoring %s, not a directory\n", dirname);
	}
}

void process(int notify, struct inotify_event* event) {
	if (event->len) {
		time_t ts;;

		struct tm* t;
		char date[20];
		int n = strlen(event->name) + strlen(dirs[event->wd - 1]);
		char filename[n+2];

		ts = time(NULL);
		t = gmtime(&ts);
		strftime(date, sizeof(date), "%F %T", t);
		out_append(date);
		out_append(" ");
		snprintf(filename, sizeof(filename), "%s/%s", dirs[event->wd - 1], event->name);
		out_append(filename);
		if (event->mask & IN_ISDIR && event->mask & IN_CREATE) {
			printf("add dir %s to watchlist\n", filename);
			add_r(notify, filename);
		}
		if (event->mask & IN_ATTRIB) { out_append(" IN_ATTRIB"); }
		if (event->mask & IN_CREATE) { out_append(" IN_CREATE"); }
		if (event->mask & IN_DELETE) { out_append(" IN_DELETE"); }
		if (event->mask & IN_DELETE_SELF) { out_append(" IN_DELETE_SELF"); }
		if (event->mask & IN_MODIFY) { out_append(" IN_MODIFY"); }
		if (event->mask & IN_MOVE_SELF) { out_append(" IN_MOVE_SELF"); }
		if (event->mask & IN_MOVED_FROM) { out_append(" IN_MOVED_FROM"); }
		if (event->mask & IN_MOVED_TO) { out_append(" IN_MOVED_TO"); }
		
		out_end_line();
	}
	alarm(MAIL_INACTIVITY_TIMEOUT);
}

void alarm_handler(int sig __attribute__((unused))) {
	flush = 1;
}

void exit_handler(int sig) {
	done = sig;
}

int main(int argc, char** argv) {
	char *dirname;
	if (argc == 1) {
		fprintf(stderr,	"Usage: %s dir_1 dir_2 ... dir_n\n",	argv[0]);
		return EXIT_FAILURE;
	}
	int notify = inotify_init();
	if (notify < 0) {
		perror("inotify_init failed");
		return EXIT_FAILURE;
	}
	

	for (int i = 1; i < argc; i++) {
		if (NULL == (dirname = canonicalize_file_name(argv[i]))) {
			perror("could not resolve path.");
			continue;
		}
		add_r(notify, dirname);
	}
	
	out_flush();

	if (SIG_ERR == signal(SIGALRM, alarm_handler)) { perror("SIGALRM");	}
	if (SIG_ERR == signal(SIGTERM, exit_handler)) { perror("SIGTERM"); }
	if (SIG_ERR == signal(SIGINT, exit_handler)) { perror("SIGINT"); }
	if (SIG_ERR == signal(SIGQUIT, exit_handler)) { perror("SIGQUIT"); }
	if (SIG_ERR == signal(SIGHUP, exit_handler)) { perror("SIGHUP"); }
	char buf[sizeof(struct inotify_event) + PATH_MAX + 10];
	int len;
	do {
		if (flush == 1) {
			puts("timout flushing...");
			out_flush();
			flush = 0;
		}
		len = read(notify, buf, sizeof(buf));

		if (len <= 0) {
			if (errno == EINTR) {
				errno = 0;
				continue;
			} else {
				perror("read failed");
				return EXIT_FAILURE;
			}
		}
		int i = 0;
		while (i < len) {
			struct inotify_event *event;
			event = (struct inotify_event *) &buf[i];
			i += sizeof(struct inotify_event) + event->len;
			process(notify, event);
		}
	} while (done == 0);
	fprintf(mail, "\nreceived signal %d. exiting...\n", done);
	pclose(mail);
	close(notify);
	free(dirname);
	return EXIT_SUCCESS;
}
